/*
 * Copyright (c) 2009, Microsoft Corporation - All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * LICENSE-GPL in the main directory of this source tree, or the
 * BSD license (http://opensource.org/licenses/bsd-license.php).
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *   Haiyang Zhang <haiyangz@microsoft.com>
 *   Hank Janssen  <hjanssen@microsoft.com>
 */
#ifndef _VMBUS_PRIVATE_H_
#define _VMBUS_PRIVATE_H_

#ifndef INTERNAL
#define INTERNAL static
#endif

#include "Hv.h"
#include "VmbusApi.h"
#include "Channel.h"
#include "ChannelMgmt.h"
#include "ChannelInterface.h"
//#include "ChannelMessages.h"
#include "RingBuffer.h"
//#include "Packet.h"
#include "List.h"
#include "timesync_ic.h"

//
// Defines
//

// Maximum channels is determined by the size of the interrupt page which is PAGE_SIZE. 1/2 of PAGE_SIZE is for
// send endpoint interrupt and the other is receive endpoint interrupt
#define MAX_NUM_CHANNELS				(PAGE_SIZE >> 1) << 3  // 16348 channels

// The value here must be in multiple of 32
// TODO: Need to make this configurable
#define MAX_NUM_CHANNELS_SUPPORTED		256

//
// Data types
//

typedef enum {
	Disconnected,
	Connecting,
	Connected,
	Disconnecting
} VMBUS_CONNECT_STATE;

#define MAX_SIZE_CHANNEL_MESSAGE			HV_MESSAGE_PAYLOAD_BYTE_COUNT

typedef struct _VMBUS_CONNECTION {

	VMBUS_CONNECT_STATE					ConnectState;

	UINT32								NextGpadlHandle;

	// Represents channel interrupts. Each bit position
	// represents a channel.
	// When a channel sends an interrupt via VMBUS, it 
	// finds its bit in the sendInterruptPage, set it and 
	// calls Hv to generate a port event. The other end
	// receives the port event and parse the recvInterruptPage
	// to see which bit is set
	VOID*								InterruptPage;
	VOID*								SendInterruptPage;
	VOID*								RecvInterruptPage;

	// 2 pages - 1st page for parent->child notification and 2nd is child->parent notification
	VOID*								MonitorPages;
	LIST_ENTRY							ChannelMsgList;
	HANDLE								ChannelMsgLock;

	// List of channels
	LIST_ENTRY							ChannelList;
	HANDLE								ChannelLock;

	HANDLE								WorkQueue;
} VMBUS_CONNECTION;


typedef struct _VMBUS_MSGINFO {
	// Bookkeeping stuff
	LIST_ENTRY			MsgListEntry;

	// Synchronize the request/response if needed
	HANDLE				WaitEvent;

	// The message itself
	unsigned char		Msg[0];
} VMBUS_MSGINFO;


//
// Externs
//
extern VMBUS_CONNECTION gVmbusConnection;

//
// General vmbus interface
//
INTERNAL DEVICE_OBJECT*
VmbusChildDeviceCreate(
	GUID deviceType,
	GUID deviceInstance,
	void *context);

INTERNAL int
VmbusChildDeviceAdd(
	DEVICE_OBJECT* Device);

INTERNAL void
VmbusChildDeviceRemove(
   DEVICE_OBJECT* Device);

//INTERNAL void
//VmbusChildDeviceDestroy(
//	DEVICE_OBJECT*);

INTERNAL VMBUS_CHANNEL*
GetChannelFromRelId(
	UINT32 relId
	);

//
// Connection interface
//
INTERNAL int
VmbusConnect(
	VOID
	);

INTERNAL int
VmbusDisconnect(
	VOID
	);

INTERNAL int
VmbusPostMessage(
	PVOID			buffer,
	SIZE_T			bufSize
	);

INTERNAL int
VmbusSetEvent(
	UINT32 childRelId
	);

INTERNAL VOID
VmbusOnEvents(
  VOID
	);


#endif // _VMBUS_PRIVATE_H_
